/*
 * Author: Nigel Kukard  <nkukard@lbsd.net>
 * Desc: Main file for BWM_GRAPH
 * Date: 26/04/2003
 * License: GPL
*/


#include <getopt.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../config.h"
#include "common.h"
#include "graph.h"
#include "misc.h"



// Function to print out our usage
void printUsage(char **argv)
{
	printf("Usage: %s [-h|--help] [options]\n\n\n",argv[0]);
	printf("Options...\n\n");
	printf("--flows=<flow1>[(counter)][:<rrd_line_type>[#rrbbgg]][,<flow2>[...]]\n");
	printf("      Flow(s) to generate report on\n\n");
	printf("--start=<start_date>\n");
    	printf("      Date/Time of report start\n\n");
	printf("--end=<end_date>\n");
	printf("      Date/Time or report end\n\n");
	printf("--output=<filename>\n");
	printf("      Output filename\n\n");
	printf("--title=<graph_title>\n");
	printf("      Title of graph\n\n\n");
	printf("Graph Options...\n\n");
	printf("--graph-avg\n");
	printf("      Add averages per flow at bottom of graph\n\n");
	printf("--graph-date\n");
	printf("      Add date information under graph\n\n");
	printf("--graph-total\n");
	printf("      Add totals per flow at bottom of graph\n\n");
}


// Main processig stuff...
int main(int argc, char **argv)
{
	int c;
	char *flows = NULL;
	// Our options
	struct option long_options[] =
	{
		{"end",1,0,'e'},
		{"flows",1,0,'f'},
		{"help",0,0,'h'},
		{"output",1,0,'o'},
		{"start",1,0,'s'},
		{"title",1,0,'t'},
		{"vertical-title",1,0,'v'},
		{"graph-avg",0,0,1},
		{"graph-total",0,0,2},
		{"graph-maxRate",0,0,3},
		{"graph-date",0,0,4},
		{0,0,0,0}
	};
	char **flowArray = NULL;
	char *outputFile = "image.png";
	int result = 0;
	char *aFlow;
	struct graphSource_t *gFlow;
	GList *graphSources = NULL;
	struct graphOptions_t *options;

	
	options = (struct graphOptions_t*) malloc0(sizeof(struct graphOptions_t));
	// FIXME - error check

	printf("BWM Graph v%s - Copyright (c) 2003 Linux Based Systems Design\n\n",PACKAGE_VERSION);

	// Print out help
	if (argc == 1)
	{
		printUsage(argv);
		return(1);
	}
	
	// Loop with options
	while (1)
	{
		int option_index = 0;
		
		// Process
		c = getopt_long(argc,argv,"e:f:hs:t:v:",long_options,&option_index);
		if (c == -1)
			break;

		// Check...
		switch (c)
		{
			case 'e':
				options->end = parseDateTime(optarg);
				if (options->end < 0)
				{
					fprintf(stderr,"ERROR: Error in end time specified\n");
					return(1);
				}
				break;
			case 'f':
				// Grab the flows...
				flows = strndup(optarg,BUFFER_SIZE);
				// Split them off
				flowArray = g_strsplit(flows,",",0);
				free(flows);
				break;
			case 'h':
				printUsage(argv);
				return(0);
			case 'o':
				outputFile = strndup(optarg,BUFFER_SIZE);
				break;
			case 's':
				options->start = parseDateTime(optarg);
				if (options->start < 0)
				{
					fprintf(stderr,"ERROR: Error in start time specified\n");
					return(1);
				}
				break;
			case 't':
				options->title = strndup(optarg,BUFFER_SIZE);
				break;
			case 'v':
				options->verticalTitle = strndup(optarg,BUFFER_SIZE);
				break;

			case 1:
				options->graph_avg = 1;
				break;
			case 2:
				options->graph_total = 1;
				break;
			case 3:
				options->graph_maxRate = 1;
				break;
			case 4:
				options->graph_date = 1;
				break;
		}
		
	}

	// Let us know about ahy unknown options
	if (optind < argc)
	{
		while (optind < argc)
			fprintf(stderr,"%s: invalid option -- %s\n",argv[0],argv[optind++]);
		return(1);
	}

	// Check everything is ok...
	if (options->start <= 0)
	{
		fprintf(stderr,"ERROR: Report start date must be specified\n");
		return(1);
	}
	if (options->end <= 0)
	{
		fprintf(stderr,"ERROR: Report end date must be specified\n");
		return(1);
	}
	if (flowArray == 0)
	{
		fprintf(stderr,"ERROR: No flows specified for reporting\n");
		return(1);
	}

	
	c = 0;
	aFlow = flowArray[c];
	while (aFlow != NULL)
	{
		// Create a proper graph flow
		gFlow = createGraphSource(aFlow);
		if (gFlow)
			// And it to our sources
			graphSources = g_list_append(graphSources,gFlow);
		// advance
		free(aFlow);
		c++;
		aFlow = flowArray[c];
	}
	free(flowArray);

	// Create our graph options...
	options->outputFile = outputFile;
	
	// Create the files we need...
	printf("Creating RRD files...\n");
	result = createRRDFiles(graphSources,options);
	
	// Write out the RRD files
	if (result == 0)
	{
		printf("Writing data to RRD files...\n");
		result = writeRRDFile(graphSources,options);
	}
	// Graph...
	if (result == 0)
	{
		printf("Graphing flows...");
		result = writeGraphs(graphSources,options);
	}
		
	return(result);
}




