/*
 * Author: Nigel Kukard  <nkukard@lbsd.net>
 * Desc: Main file for BWMD
 * Date: 18/04/2003
 * License: GPL
*/


#include <getopt.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "../config.h"
#include "common.h"
#include "flow.h"
#include "ipq.h"
#include "report.h"
#include "flowControl.h"
#include "xmlConf.h"


void modprobe(gpointer key, gpointer value, gpointer user_data)
{
	char buffer[4096];
	int *outcome = (int *) user_data;
	int ret;

		
	// Generate the command we need
	if (value)
	{
		snprintf(buffer,4096,"/sbin/modprobe %s %s > /dev/null 2>&1", (char *) key, (char *) value);
		fprintf(stderr,"Loading %s with \"%s\"...", (char *) key, (char *) value);
	}
	else
	{
		snprintf(buffer,4096,"/sbin/modprobe %s > /dev/null 2>&1", (char *) key);
		fprintf(stderr,"Loading %s...", (char *) key);
	}
	// Run it and check
	ret = system(buffer);
	if (WEXITSTATUS(ret) != 0)
	{
		fprintf(stderr,"failed\n");
		*outcome = 0;
	}
	else
		fprintf(stderr,"done\n");
}


// Main processig stuff...
int main(int argc, char **argv)
{
	struct runnerData_t *runnerData;
	GThread *IPQRunnerThread;
	GThread *flowRunnerThread;
	GThread *controlRunnerThread;
	GThread *reportRunnerThread;
	int daemonize = 1;
	int c, i;
	char *configFile = CONFIG_FILE;
	struct flowData_t flowData;
	GHashTable *moduleHash; //, *moduleItem;


	// Our options
	struct option long_options[] =
	{
		{"config",1,0,'c'},
		{"foreground",0,0,'f'},
		{"help",0,0,'h'},
		{0,0,0,0}
	};
	
	
	printf("BWM Daemon v%s - Copyright (c) 2003 Linux Based Systems Design\n\n",PACKAGE_VERSION);
	
	// Loop with options
	while (1)
	{
		int option_index = 0;
		
		// Process
		c = getopt_long(argc,argv,"c:fh",long_options,&option_index);
		if (c == -1)
			break;

		// Check...
		switch (c)
		{
			case 'c':
				configFile = strdup(optarg);
				break;
			case 'f':
				daemonize = 0;
				break;
			case 'h':
				// FIXME - version from CVS
				printf("Usage: %s [-h|--help] [--config=<config_file>] [--foreground]\n",argv[0]);
				return(0);
		}
		
	}

	// Let us know about ahy unknown options
	if (optind < argc)
	{
		while (optind < argc)
			fprintf(stderr,"%s: invalid option -- %s\n",argv[0],argv[optind++]);
		return(1);
	}


	// Allocate some memory
	runnerData = (struct runnerData_t*) malloc0(sizeof(struct runnerData_t));
				
	// Initialize thread system
	g_thread_init(NULL);
	if (!g_thread_supported())
	{
		fprintf(stderr,"Failed to initialize thread system\n");
		return(2);
	}


	// Grab our flows...	
	flowData = createFlowData(configFile);
	if (!flowData.flows)
	{
		fprintf(stderr,"Failed to load flow data from configuration file!\n");
		return(3);
	}


	if (!daemonize)
		printf("BWMD: Loaded %i flows and %i queues\n",g_list_length(flowData.flows),g_list_length(flowData.pktQueues));

	
	// Time to load our modules
	moduleHash = getModuleLoadHash(configFile);
	if (!moduleHash)
	{
		fprintf(stderr,"Failed to load module list\n");
		return(4);
	}

	if (!daemonize)
		printf("BWMD: Found %i modules to load\n",g_hash_table_size(moduleHash));


	i = 1;
	g_hash_table_foreach(moduleHash,modprobe,&i);
	if (!i)
	{
		fprintf(stderr,"Failed to modprobe module(s)\n");
		return(5);
	}
	g_hash_table_destroy(moduleHash);
	
	// Setup the queue runners'd data
	runnerData->daemon = daemonize;

	runnerData->pktQueues = flowData.pktQueues;
	runnerData->flows = flowData.flows;
	runnerData->groups = flowData.groups;
	
	runnerData->bandSignalLock = g_mutex_new();
	runnerData->bandSignalCond = g_cond_new();
	runnerData->queueChanged = 0;
	runnerData->waitListLock = g_mutex_new();
	runnerData->queueWait = 0;

	// Initialize the arrays
	for (i = 0; i < NUM_PRIO_BANDS; i++)
	{
		runnerData->queueChangeList[i] = NULL;
		runnerData->queueWaitList[i] = NULL;
	}

	runnerData->pmem.freeList = NULL;
	runnerData->pmem.lock = g_mutex_new();

	runnerData->IPQLock = g_mutex_new();

	// Check if we must become a daemon
	if (daemonize)
		daemon(0,0);
	
	// Create our threads to run
	IPQRunnerThread = g_thread_create(IPQRunner, (void *) runnerData, TRUE, NULL);
	flowRunnerThread = g_thread_create(flowRunner, (void *) runnerData, TRUE, NULL);
	controlRunnerThread = g_thread_create(controlRunner, (void *) runnerData, TRUE, NULL);
	reportRunnerThread = g_thread_create(reportRunner, (void *) runnerData, TRUE, NULL);


	g_thread_join(IPQRunnerThread);
	
	
	return 0;
}


