/*
 * Author: Nigel Kukard  <nkukard@lbsd.net>
 * Desc: flow.h header file for bwmd
 * Date: 15/04/2003
 * License: GPL
*/


#ifndef _FLOW_H
#define _FLOW_H


#include <glib.h>
#include <linux/netfilter.h>
#include <sys/time.h>
#include "common.h"
#include "ipq.h"
#include "libipq.h"


// Parent flow member,  yyyy = member
#define P_FLOW(pktQueue,yyyy) (pktQueue->parentFlow)->yyyy

// Packet payload length
#define PKT_SIZE(packet) packet->payload->data_len
// Packet id
#define PKT_ID(packet) packet->payload->packet_id

// QUEUE exceeded?
#define QUEUE_EXCEEDED(pktQueue) \
	((pktQueue->curLen >= pktQueue->maxLen && pktQueue->maxLen != 0) || \
	(pktQueue->curSize > pktQueue->maxSize && pktQueue->maxSize != 0))
// Credit exceeded?
#define CREDIT_EXCEEDED(flow,packet,maxCredit,burstCredit) \
	((maxCredit - PKT_SIZE(packet) < 0 && flow->maxRate > 0 && flow->burstRate == -1) || \
	(burstCredit - PKT_SIZE(packet) < 0 && flow->burstRate > 0 && flow->maxRate != 0))
// FLOW queue exceeded?
#define FLOW_QUEUE_EXCEEDED(flow) \
	((flow->curQueueLen >= flow->maxQueueLen && flow->maxQueueLen != 0) || \
	(flow->curQueueSize > flow->maxQueueSize && flow->maxQueueSize != 0))
// Threshold exceeded?
#define TH_EXCEEDED(pktQueue,threshold) \
	(P_FLOW(pktQueue,curQueueLen) > P_FLOW(pktQueue,threshold) && P_FLOW(pktQueue,maxQueueLen) != 0)
// Get new credit
#define CALC_MAX_CREDIT(curTime,flow,creditVariable) \
			if (flow->maxRate > 0) \
			{ \
				/* Calculate credit */ \
				if (curTime.tv_sec - flow->lastPacketTime.tv_sec > 0) \
					creditVariable = flow->maxRate; \
				else \
					creditVariable = flow->curMaxCredit + ((curTime.tv_usec - flow->lastPacketTime.tv_usec) * flow->usMaxCredit); \
				/* Check if we exceed */ \
				if (creditVariable > flow->maxRate) \
					creditVariable = flow->maxRate; \
			}
#define CALC_BURST_CREDIT(curTime,flow,creditVariable) \
			if (flow->burstRate > 0) \
			{ \
				/* Calculate credit */ \
				if (curTime.tv_sec - flow->lastPacketTime.tv_sec > 0) \
					creditVariable = flow->burstRate; \
				else \
					creditVariable = flow->curBurstCredit + ((curTime.tv_usec - flow->lastPacketTime.tv_usec) * flow->usBurstCredit); \
				/* Check if we exceed */ \
				if (creditVariable > flow->burstRate) \
					creditVariable = flow->burstRate; \
			}


// Our packet statistic structure
struct pktStat_t
{
	unsigned int pktCount;
	unsigned int pktSize;
	unsigned int pktDropped;
	unsigned int pktBursted;
};


// A packet queue
struct pktQueue_t
{
	GMutex *lock;

	long int prio;
	
	struct flow_t *parentFlow;  // void* because we cannot ref below?

	long int nfmark;
	
	long int curSize;
	long int curLen;
	long int maxSize;
	long int maxLen;

	GList *packets;
};

// A flow
struct flow_t
{
	// Lock is used to lock basically everything, cept counter
	GMutex *lock;

	// Flow identifier
	char flowName[MAX_REFLEN];  // STATIC - my reference

	// Stats stiff
	long int statsLen;  // STATIC - max length of stat array
	long int statsPos;
	struct pktStat_t *pktStats;  // uses ++ to increase in temp var to get to statsLen

	// Shaping	
	struct flow_t *parent;  // STATIC, not specified, determined - pointer to parent
	int prio;  // STATIC - Flow priority
	long int nfmark;  // STATIC - nfmark value i must match auto to queues
	long int maxQueueSize;  // STATIC - max length in bytes of queue
	long int maxQueueLen;  // STATIC - and/or length in items
	long int maxRate;  // STATIC - max rate in bytes per second
	long int burstRate;  // STATIC - max rate bytes per second
	long int curQueueSize;  // Current queue length in bytes, this is included in the runningLock
	long int curQueueLen;  // Current queue length in items, this is included in the runningLock
	struct pktQueue_t *pktQueues[NUM_PRIO_BANDS];

	// Setup our thresholds
	long int min_th;  // We start dropping packets here
	long int max_th;  // We drop alot more here
	float parent_th;  // Used to check the available percentage of parent bandwidth before we burst
	
	// Counters, used for both shape & counter logging
	GMutex *counterLock;  // Lock used only for locking these counters
	struct pktStat_t counter;  // This is the counter which is logged to file
	GList *groups;

	float usMaxCredit;  // Credits for maxRate per microsecond
	float curMaxCredit;  // Current available credit for maxRate      0 <= curCredit <= maxRate
	float usBurstCredit;
	float curBurstCredit;
	struct timeval lastPacketTime;  // Last "timeval" that we had a packet come in
	struct pktStat_t running;  // This is the running counter
	
	int lastDumpTimestamp;
	int counterTimeout;  // STATIC - length between file logs
	int counterRemaining;
};

// A group
struct group_t
{
	// Lock is used to lock basically everything, cept counter
	GMutex *lock;

	// Group identifier
	char groupName[MAX_REFLEN];

	// Stats stiff
	long int statsLen;  // STATIC - max length of stat array
	long int statsPos;
	struct pktStat_t *pktStats;  // uses ++ to increase in temp var to get to statsLen

	// Counters
	GMutex *counterLock;  // Lock used only for locking these counters
	struct pktStat_t counter;
	struct pktStat_t running;

	int lastDumpTimestamp;
	int counterTimeout;  // STATIC - length between file logs
	int counterRemaining;

	// Flows we "contain"
	GList *flowList;
};

// Get avg flow stats
struct pktStat_t getFlowTotalStats(struct pktStat_t *pktStats, long int pktStatsLen);

// Create a flow
struct flow_t* createFlow(
		char *flowName,
		struct flow_t *parentFlow,
		long int statsLen, 
		long int maxQueueSize, 
		long int maxQueueLen, 
		long int maxRate,
		long int burstRate,
		long int nfmark,
		float parent_th,
		int reportTimeout);

// Create group
struct group_t *createGroup(
		char *groupName, 
		long int statsLen, 
		int reportTimeout, 
		GList *flowList);

// Function to update all our flow groups
void updateGroups(
		struct flow_t *flow, 
		int pktCount, 
		int pktSize, 
		int pktDropped, 
		int pktBursted);
	
// Create a packet queue
struct pktQueue_t *createPktQueue(
		long int prio,
		long int nfmark,
		struct flow_t *parentFlow);

	
// Flow runner thread
void *flowRunner(void *data);


#endif

