/*
 * Author: Nigel Kukard  <nkukard@lbsd.net>
 * Desc: Flow handling for bwmd
 * Date: 15/04/2003
 * License: GPL
*/


#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "common.h"
#include "flow.h"
#include "ipq.h"





// Create a flow
struct flow_t* createFlow(
		char *flowName,
		struct flow_t *parentFlow,
		long int statsLen, 
		long int maxQueueSize, 
		long int maxQueueLen, 
		long int maxRate,
		long int burstRate,
		long int nfmark,
		float parent_th,
		int reportTimeout)
{
	int i;
	struct flow_t *flow = (struct flow_t*) malloc(sizeof(struct flow_t));


	// Check if its ok...
	if (!flow)
	{
		// FIXME - log error
		fprintf(stderr,"Failed to allocate memory for flow\n");
		return NULL;
	}

	// Lock entire flow
	flow->lock = g_mutex_new();
	
	strncpy(flow->flowName,flowName,MAX_REFLEN); // copy string
	
	flow->statsLen = statsLen;
	flow->statsPos = 0;

	flow->parent = parentFlow;
	flow->maxQueueSize = maxQueueSize;
	flow->maxQueueLen = maxQueueLen;
	flow->maxRate = maxRate;
	flow->burstRate = burstRate;
	flow->nfmark = nfmark;

	flow->parent_th = parent_th;
	
	flow->counterTimeout = reportTimeout;
	flow->counterRemaining = reportTimeout;

	flow->curQueueSize = 0;
	flow->curQueueLen = 0;

	// Lock for our counters only
	flow->counterLock = g_mutex_new();
	flow->counter.pktCount = 0;
	flow->counter.pktSize = 0;
	flow->counter.pktDropped = 0;
	flow->counter.pktBursted = 0;

	flow->groups = NULL;
	
	flow->lastDumpTimestamp = time(NULL);
	// Check if we can set our credit
	if (flow->maxRate > 0)
	{
		flow->usMaxCredit = flow->maxRate / 1000000.0;  // For micro (millionths) of seconds silly
		flow->curMaxCredit = flow->maxRate;
	}
	else
	{
		flow->usMaxCredit = 0;
		flow->curMaxCredit = 0;
	}
	
	if (flow->burstRate > 0)
	{
		flow->usBurstCredit = flow->burstRate / 1000000.0;	
		flow->curBurstCredit = flow->burstRate;
	}
	else
	{
		flow->usBurstCredit = 0;
		flow->curBurstCredit = 0;
	}
	// Set last packet time to now
	gettimeofday(&flow->lastPacketTime,NULL);
	// Clear our running counters
	flow->running.pktCount = 0;
	flow->running.pktSize = 0;
	flow->running.pktDropped = 0;
	flow->running.pktBursted = 0;

	// Verify everything is ok...
	if (flow->statsLen == 0)
		flow->statsLen = 10;
	
	// Create stuff...
	flow->pktStats = (struct pktStat_t*) malloc0(sizeof(struct pktStat_t) * flow->statsLen);
	if (!flow->pktStats)
	{
		// FIXME - log error
		fprintf(stderr,"Failed to allocate memory for flow->pktStats\n");
		return NULL;
	}

	// Calculate the queue length ... etc
	if (flow->maxQueueLen == -1)
	{
		if (flow->maxRate != 0)
			flow->maxQueueLen = (flow->maxRate / 750) * 5;  // Half packet size (how many half packets we can store) in 5s
		else
			flow->maxQueueLen = 0;
	}
	if (flow->maxQueueSize == -1)
	{
		if (flow->maxRate != 0)
			flow->maxQueueSize = flow->maxRate * 5;  // Let us queue a 5 seconds of packets
		else
			flow->maxQueueSize = 0;
	}

	// Calculate our thresholds
	flow->min_th = (flow->maxRate / 750);
	flow->max_th = (flow->maxRate / 750) * 4;

	// Blank all queues
	for (i = 0; i < NUM_PRIO_BANDS; i++) 
	{
		struct pktQueue_t *pktQueue = flow->pktQueues[i] = (struct pktQueue_t*) malloc(sizeof(struct pktQueue_t));
		
		pktQueue->lock = g_mutex_new();
		pktQueue->prio = i;
		pktQueue->parentFlow = flow;
		pktQueue->nfmark = -1;
		pktQueue->curSize = 0;
		pktQueue->curLen = 0;
		pktQueue->maxSize = 0;
		pktQueue->maxLen = 0;
		pktQueue->packets = NULL;
	}

/*
	printf("Flow added...\n");
	printf("    flowName     - %s\n",flowName);
	printf("    parent       - %p\n",parentFlow);
	printf("    nfmark       - %li\n",nfmark);
	printf("    statsLen     - %li\n",flow->statsLen);
	printf("    maxQueueSize - %li\n",flow->maxQueueSize);
	printf("    maxQueueLen  - %li\n",flow->maxQueueLen);
	printf("    maxRate      - %li\n",flow->maxRate);
	printf("    usMaxCredit  - %.6f\n",flow->usMaxCredit);
	printf("    curMaxCredit - %.6f\n",flow->curMaxCredit);
	printf("    usCredit     - %.6f\n",flow->usBurstCredit);
	printf("    curCredit    - %.6f\n",flow->curBurstCredit);
	printf("    burstRate    - %li\n",flow->burstRate);
	printf("Queue Stuff:\n");
	printf("    min_th       - %li\n",flow->min_th);
	printf("    max_th       - %li\n",flow->max_th);
*/
	
	return flow;	
}


// Create group
struct group_t *createGroup(
		char *groupName, 
		long int statsLen, 
		int reportTimeout, 
		GList *flowList)
{
	struct group_t *group = (struct group_t *) malloc(sizeof(struct group_t));

	
	// Lock entire flow
	group->lock = g_mutex_new();
	
	strncpy(group->groupName,groupName,MAX_REFLEN); // copy string
	
	// Stats stuff
	group->statsLen = statsLen;
	group->statsPos = 0;
	
	// Clear our running counters
	group->running.pktCount = 0;
	group->running.pktSize = 0;
	group->running.pktDropped = 0;
	group->running.pktBursted = 0;
	
	group->counterLock = g_mutex_new();
	group->counter.pktCount = 0;
	group->counter.pktSize = 0;
	group->counter.pktDropped = 0;
	group->counter.pktBursted = 0;

	// Set dump time
	group->lastDumpTimestamp = time(NULL);

	// Report stuff
	group->counterTimeout = reportTimeout;
	group->counterRemaining = reportTimeout;

	// Our list of flows
	group->flowList = flowList;
	
	// Verify everything is ok...
	if (group->statsLen == 0)
		group->statsLen = 10;

	group->pktStats = (struct pktStat_t*) malloc0(sizeof(struct pktStat_t) * group->statsLen);
	if (!group->pktStats)
	{
		// FIXME - log error
		fprintf(stderr,"Failed to allocate memory for group->pktStats\n");
		return NULL;
	}
	return group;
}


// Create a packet queue for the flow
struct pktQueue_t *createPktQueue(
		long int prio,
		long int nfmark,
		struct flow_t *parentFlow)
{
	struct pktQueue_t *pktQueue = parentFlow->pktQueues[prio];

	
	// Make sure we valid
	if (prio >= NUM_PRIO_BANDS || prio < 0)
	{
		fprintf(stderr,"ERROR: Queue priority must be between 0 and 99\n");
		return NULL;
	}

	pktQueue->nfmark = nfmark;
/*	
	printf("  Queue added...\n");
	printf("      prio     - %li\n",prio);
	printf("      nfmark   - %li\n",nfmark);
*/	
	return pktQueue;	
}


// Function to update all our flow groups
void updateGroups(struct flow_t *flow, int pktCount, int pktSize, int pktDropped, int pktBursted)
{
	GList *groupItem = g_list_first(flow->groups);


	// Loop while we have summin
	while (groupItem)
	{
		struct group_t *tmpGroup = groupItem->data;


		// Lock and load!
		g_mutex_lock(tmpGroup->lock);

		// Do our stuff
		if (pktCount == -1)
			tmpGroup->running.pktCount = 0;
		else
			tmpGroup->running.pktCount += pktCount;

		if (pktSize == -1)
			tmpGroup->running.pktSize = 0;
		else
			tmpGroup->running.pktSize += pktSize;

		if (pktDropped == -1)
			tmpGroup->running.pktDropped = 0;
		else
			tmpGroup->running.pktDropped += pktDropped;

		if (pktBursted == -1)
			tmpGroup->running.pktBursted = 0;
		else
			tmpGroup->running.pktBursted += pktBursted;
			
		g_mutex_unlock(tmpGroup->lock);

		// Next one plz	
		groupItem = g_list_next(groupItem);
	}
}


// Get average stats len, make sure you lock the flow!!!
struct pktStat_t getFlowTotalStats(struct pktStat_t *pktStats, long int pktStatsLen)
{
	struct pktStat_t returnStats;
	long int curStat = 0;

	
	// Reset return values
	returnStats.pktCount = 0;
	returnStats.pktSize = 0;
	returnStats.pktDropped = 0;
	returnStats.pktBursted = 0;
	// Gather stats	
	for (curStat = 0; curStat < pktStatsLen; curStat++)
	{
		returnStats.pktCount += pktStats[curStat].pktCount;
		returnStats.pktSize += pktStats[curStat].pktSize;
		returnStats.pktDropped += pktStats[curStat].pktDropped;
		returnStats.pktBursted += pktStats[curStat].pktBursted;
	}

	// And return
	return returnStats;	
}



// Function to process a flow queue, returns number of packets accepted
static int processPktQueue(struct runnerData_t *runnerData, struct pktQueue_t *pktQueue)
{
	int status;
	GList *packets;
	int exceeded = 0;
	int i;
	// Differences in queue when we done
	int acceptLen = 0, queuedLen = 0;
	int acceptSize = 0, queuedSize = 0;


	// Lock, hijack packets, unlock
	g_mutex_lock(pktQueue->lock);
	packets = pktQueue->packets;
	pktQueue->packets = NULL;
	g_mutex_unlock(pktQueue->lock);

	// Check that we within our boundaries	
	while (!exceeded)
	{
		struct packet_t *packet;
		GList *pktQueueItem;
		int ok = 1;
		struct flow_t *flow;
		struct flow_t *childFlow = NULL;


		// Get head of the queue
		pktQueueItem = g_list_first(packets);
		// Last item
		if (!pktQueueItem)
			break;
		// We got summin
		packet = pktQueueItem->data;
	
		
		childFlow = flow = pktQueue->parentFlow;
		// Loop to root
		while (flow && ok)
		{
			struct timeval timeNow;
			float maxCredit = 0, burstCredit = 0;
			
		
			// Grab exact time now
			gettimeofday(&timeNow,NULL);

			// Set new credits
			CALC_MAX_CREDIT(timeNow,flow,maxCredit)
			CALC_BURST_CREDIT(timeNow,flow,burstCredit)
			
			// Check if we have exceeded stuff we shouldn't	
			if (CREDIT_EXCEEDED(flow,packet,maxCredit,burstCredit))
				ok = 0;
			// If we have a parent, if we do... do a few tests
			if (ok && flow->parent)
			{
				// Check if we havn't exceeded our parent queues
				if (FLOW_QUEUE_EXCEEDED(flow->parent))
					ok = 0;

				// Check if we have a bogged parent that we cannot burst to
				// FIXME: Add a percentage of traffic in parent that is free so we can burst
				// FIXME: What about parent burst rates?
				if (ok && (maxCredit - PKT_SIZE(packet) < 0 && flow->maxRate > 0))
				{
					float parentMax = 0, parentBurst = 0;


					// Calculate parent credits
					CALC_MAX_CREDIT(timeNow,flow->parent, parentMax)
					CALC_BURST_CREDIT(timeNow,flow->parent, parentBurst)

					// Check if we exceeded
					if (CREDIT_EXCEEDED(flow->parent,packet,parentMax,parentBurst) ||
							((flow->parent_th > ((float) parentMax / (float) flow->parent->maxRate) * 100.0) && 
							flow->parent_th > 0))
						ok = 0;	
				}
			}
			// If packet is still ok to pass through, do our stuff	
			if (ok)
			{
				g_mutex_lock(flow->lock);
			
				flow->running.pktCount++;
				flow->running.pktSize += PKT_SIZE(packet);
				flow->curMaxCredit = maxCredit - PKT_SIZE(packet);
				flow->curBurstCredit = burstCredit - PKT_SIZE(packet);

				// We bursted
				if (maxCredit < 0 && flow->maxRate != 0)
					flow->running.pktBursted++;
				// Last time we accepted
				gettimeofday(&flow->lastPacketTime,NULL);
				
				g_mutex_unlock(flow->lock);
		
				// Time to update our groups
				updateGroups(flow,1,PKT_SIZE(packet),0,0);
				
				// This flow is inbetween, we accepting traffic through, counters must be updated below
				childFlow = flow;
				flow = flow->parent;
			}
		}
		

		// Can we accept the packet
		if (ok)
		{
			// Re-inject packet for processing, check if it was ok...
			g_mutex_lock(runnerData->IPQLock);
			status = ipq_set_verdict(runnerData->IPQHandle, PKT_ID(packet), NF_ACCEPT, 0, NULL);
			g_mutex_unlock(runnerData->IPQLock);
			if (status < 0)
			{
				// FIXME - send error to log
				fprintf(stderr,"Failed to ACCEPT packet\n");
				break;
			}

			// Add to our flow counters, remove from queue and free
			acceptSize += PKT_SIZE(packet);
			acceptLen++;
			packets = g_list_remove(packets,packet);
			// Save the size
			i = PKT_SIZE(packet);
			// FIXME - UPDATE GROUPS
			// "Free" the packet
			g_mutex_lock(runnerData->pmem.lock);
			runnerData->pmem.freeList = g_list_append(runnerData->pmem.freeList,packet);
			g_mutex_unlock(runnerData->pmem.lock);
		}
		else
		{
			// We got no where
			if (flow == childFlow)
				exceeded = 1;
			else
			{
				struct pktQueue_t *nextPktQueue = flow->pktQueues[pktQueue->prio];


				fprintf(stderr,"Overflowing to %s...\n",flow->flowName);
				// We can do this before locking
				queuedLen++;
				queuedSize += PKT_SIZE(packet);
				packets = g_list_remove(packets,packet);
			
				// Lock everything
				g_mutex_lock(nextPktQueue->lock);
				g_mutex_lock(flow->lock);
				// Add packet to parent queue & remove from child
				nextPktQueue->packets = g_list_append(nextPktQueue->packets,packet);
				// Update next queue stats	
				nextPktQueue->curLen++;
				nextPktQueue->curSize += PKT_SIZE(packet);
				// Update next flow stats
				flow->curQueueLen++;
				flow->curQueueSize += PKT_SIZE(packet);
				// Finally unlock	
				g_mutex_unlock(flow->lock);
				g_mutex_unlock(nextPktQueue->lock);

				// Signal that we just added to the queue
				g_mutex_lock(runnerData->bandSignalLock);
				// Check if we havn't already gotten the queue listed
				// FIXME: check if we can't just add, high flow queue
				if (!g_list_find(runnerData->queueChangeList[nextPktQueue->prio],nextPktQueue))
					runnerData->queueChangeList[nextPktQueue->prio] = 
						g_list_append(runnerData->queueChangeList[nextPktQueue->prio],nextPktQueue);
				g_cond_signal(runnerData->bandSignalCond);
				g_mutex_unlock(runnerData->bandSignalLock);
			}

		}
	}

	// Lock our data and change it, insert left over packets aswell
	g_mutex_lock(P_FLOW(pktQueue,lock));
	P_FLOW(pktQueue,curQueueLen) -= (acceptLen + queuedLen);
	P_FLOW(pktQueue,curQueueSize) -= (acceptSize + queuedSize);
	g_mutex_unlock(P_FLOW(pktQueue,lock));

	// Update queue aswell
	g_mutex_lock(pktQueue->lock);
	if (packets)
	{
		pktQueue->packets = g_list_concat(packets,pktQueue->packets);
		// Signal that we just added to the queue
		g_mutex_lock(runnerData->bandSignalLock);
		// Check if we havn't already gotten the queue listed
		// FIXME: check if we can't just add, high flow queue
		if (!g_list_find(runnerData->queueChangeList[pktQueue->prio],pktQueue))
			runnerData->queueChangeList[pktQueue->prio] = g_list_append(runnerData->queueChangeList[pktQueue->prio],pktQueue);
		g_cond_signal(runnerData->bandSignalCond);
		g_mutex_unlock(runnerData->bandSignalLock);
	}
	pktQueue->curLen -= (acceptLen + queuedLen);
	pktQueue->curSize -= (acceptSize + queuedLen);

	g_mutex_unlock(pktQueue->lock);

				// Quickly and update everything in between
//FIXME!!!!!				updateGroups(runnerData,flow,1,i,0,0);

	return(acceptLen);
}



// Main IPQ->QUEUE thread
void *flowRunner(void *data)
{
	struct runnerData_t *runnerData = (struct runnerData_t*) data;
	GList *queueChangeList[NUM_PRIO_BANDS];
	int i, pktsProcessed = 0;
	GTimeVal mytime;


	// Our processing function
	void processQueue(void *data, void *user_data)
	{
		struct pktQueue_t *pktQueue = (struct pktQueue_t*) data;
		struct runnerData_t *aRunnerData = (struct runnerData_t*) user_data;


		// Check if we found a flow with a queue
		pktsProcessed += processPktQueue(aRunnerData,pktQueue);
//		printf("%i packets processed\n",pktsProcessed);
	}

	
	// Allocate queue change list
	for (i = 0; i < NUM_PRIO_BANDS; i++)
		queueChangeList[i] = NULL;

	
	if (!runnerData->daemon)
		printf("Flow runner started...\n");

	// Loop when we get a signal
	while (1)
	{
		g_mutex_lock(runnerData->bandSignalLock);

		// If the queue has changed proceed
		if (runnerData->queueChanged == 0)
		{
			g_get_current_time(&mytime);
			g_time_val_add(&mytime,10000);
			g_cond_timed_wait(runnerData->bandSignalCond,runnerData->bandSignalLock,&mytime);
		}
		
		
		// Hijack the queue change list items
		for (i = 0; i < NUM_PRIO_BANDS; i++)
		{
			// Copy list item over
			queueChangeList[i] = runnerData->queueChangeList[i];
			// Zero runner data if it is non-NULL
			if (runnerData->queueChangeList[i])
			{
			       runnerData->queueChangeList[i] = NULL;
			}
			else
				queueChangeList[i] = NULL;
		}
		
		runnerData->queueChanged = 0;

		g_mutex_unlock(runnerData->bandSignalLock);

		
		// Process all list
		for (i = 0; i < NUM_PRIO_BANDS; i++)
		{
			// Process list if it is non-NULL
			if (queueChangeList[i])
			{
				g_list_foreach(queueChangeList[i],processQueue,runnerData);
				g_list_free(queueChangeList[i]);
			}
		}
	}

	return(NULL);
}

	
