/*
 * Author: Nigel Kukard  <nkukard@lbsd.net>
 * Desc: XML config file stuff...
 * Date: 05/05/2003
 * License: GPL
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include "common.h"
#include "flow.h"
#include "xmlConf.h"



/* general functions - static struct confACLChain_t *lookupChain(GHashChain *tables, char *tableName) {{{ */
// Function to get a chain if it exists or to create it
static struct confACLChain_t *lookupChain(GHashTable *chains, char *chainName)
{
	struct confACLChain_t *tmpChain;
	
	// See if we have a table by this name
	tmpChain = g_hash_table_lookup(chains,chainName);
	if (tmpChain == NULL)
	{	
		// If not... create everything
		tmpChain = (struct confACLChain_t*) malloc(sizeof(struct confACLChain_t));
		tmpChain->name = strdup(chainName);
		tmpChain->defaultTarget = NULL;
		// If not... start creating everything...
		tmpChain->ruleList = NULL;
		tmpChain->ruleset = NULL;
		g_hash_table_insert(chains,chainName,tmpChain);
	}

	// Return chain
	return(tmpChain);
}
/* }}} */

/* general functions - static struct confACLTable_t *lookupTable(GHashTable *tables, char *tableName) {{{ */
// Function to get a table if it exists or to create it
static struct confACLTable_t *lookupTable(GHashTable *tables, char *tableName)
{
	struct confACLTable_t *tmpTable;
	
	// See if we have a table by this name
	tmpTable = g_hash_table_lookup(tables,tableName);
	if (tmpTable == NULL)
	{	
		// If not... create everything
		tmpTable = (struct confACLTable_t*) malloc(sizeof(struct confACLTable_t));
		tmpTable->name = strdup(tableName);
		tmpTable->chains = g_hash_table_new(g_str_hash,g_str_equal);
		g_hash_table_insert(tables,tableName,tmpTable);
	}

	// Return table
	return(tmpTable);
}
/* }}} */

/* general functions - static GList *splitContents(xmlNodePtr node) {{{ */
// Function to split out the contents of an xml content into a linked list
static GList *splitContents(xmlNodePtr node)
{
	char *contents = xmlNodeGetContent(node);
	char **contentArray;
	int i,j,k;
	char *aLine;
	char *bLine;
	GList *result = NULL;


	// Alloc sum memory
	bLine = (char *) malloc(BUFFER_SIZE);
	 
	// Split off contents per line
	contentArray = g_strsplit(contents,"\n",0);
	
	// Loop with all contents
	i = 0;
	aLine = contentArray[i];
	while (aLine)
	{
		// Build bLine out of valid chars	
		memset(bLine,'\0',BUFFER_SIZE);
		k = 0;
		for (j = 0; j < strlen(aLine); j++)
		{
			// Check for valid chars
			if ((aLine[j] >= 48 && aLine[j] <= 57) || (aLine[j] >= 65 && aLine[j] <= 90) || 
					aLine[j] == 95 || (aLine[j] >= 97 && aLine[j] <= 122) || aLine[j] == 124)
			{
				bLine[k] = aLine[j];
				k++;
			}
		}	
		// If we actually got something add it to the list
		if (strlen(bLine) > 0)
		{
			result = g_list_append(result,bLine);
			bLine = (char *) malloc(BUFFER_SIZE);
		}

		// Carry on...
		i++;
		aLine = contentArray[i];
	}

	// Make sure we free the last one
	free(bLine);
	
	return(result);
}
/* }}} */

/* general functions - static GHashTable *getProperties(xmlNodePtr node) {{{ */
// Function to get the properties in a hash table of a node
static GHashTable *getProperties(xmlNodePtr node)
{
	xmlAttrPtr attrib;
	xmlChar	*attrName;
	xmlChar *contentNode;
	GHashTable *attribHash;
	
	
	attribHash = g_hash_table_new(g_str_hash,g_str_equal);

	// Get an easy to use pointer to our properties and iterate
	attrib = node->properties;
	while (attrib)
	{
		// Save our property to our hash	
		attrName = (xmlChar *) attrib->name;
		contentNode = ((xmlNodePtr) attrib->children)->content;
		g_hash_table_insert(attribHash,attrName,strdup(contentNode));
		// Advance property
		attrib = attrib->next;
	}

	// Return our hash
	return(attribHash);
}
/* }}} */

/* ruleset functions - static GList *createRuleset(GHashTable *classHash, GList *classList, char *chainName, char *target, char *cmd_line1, char *cmd_line2) {{{ */
// Function to build a ruleset out of a number of classes and an extra char* at the end
static GList *createRuleset(GHashTable *classHash, GList *classList, char *chainName, char *target, char *cmd_line1, char *cmd_line2)
{
	char *className;
	GList *result = NULL;
	int found;
	
	
	// Function to process ruleset, and append to result
	void processRuleset(struct confClass_t *aClass)
	{
		// Function to append to return ruleset
		void processRulesetItem(gpointer data, gpointer user_data)
		{
			char *rule = (char *) data;
			char *buffer = (char *) malloc0(BUFFER_SIZE);
			int done = 0;

			
			// If we have all
			if (chainName && rule && target && cmd_line1 && cmd_line2)
			{
				snprintf(buffer,BUFFER_SIZE,"-A %s %s %s -j %s %s",chainName,rule,cmd_line1,target,cmd_line2);
				done = 1;
			}
			if (chainName && rule && target && cmd_line1)
			{
				snprintf(buffer,BUFFER_SIZE,"-A %s %s %s -j %s",chainName,rule,cmd_line1,target);
				done = 1;
			}
			if (!done && chainName && rule && target && cmd_line2)
			{
				snprintf(buffer,BUFFER_SIZE,"-A %s %s -j %s %s",chainName,rule,target,cmd_line2);
				done = 1;
			}
			if (!done && chainName && rule && target)
			{
				snprintf(buffer,BUFFER_SIZE,"-A %s %s -j %s",chainName,rule,target);
				done = 1;
			}
			if (!done && cmd_line1)
			{
				snprintf(buffer,BUFFER_SIZE,"%s",cmd_line1);
				done = 1;
			}
			if (!done && cmd_line2)
			{
				snprintf(buffer,BUFFER_SIZE,"%s",cmd_line2);
				done = 1;
			}
	
			// Free buffer if we don't need it	
			if (done)
				result = g_list_append(result,buffer);
			else
				free(buffer);
		}

		// Loop with ruleset items
		g_list_foreach(aClass->ruleset,processRulesetItem,NULL);
	}

	// Function to loop with all classes and find the one we need
	void processClass(gpointer data, gpointer user_data)
	{
		// Function to find a class by name
		void processClassHash(gpointer p_key, gpointer p_value, gpointer user_data)
		{
			char *key = (char *) p_key;
			struct confClass_t *value = (struct confClass_t *) p_value;

			// Check if we found our class
			if (!found)
				if (!xmlStrcmp(key,className))
				{
					processRuleset(value);
					found = 1;
				}
		}
	

		className = (char *) data;
		found = 0;
		// Loop with class hash
		g_hash_table_foreach(classHash,processClassHash,NULL);
		// Check if we found it
		if (!found)
			fprintf(stderr,"ERROR: Class %s invalid\n",className);
	}


	// Loop through all our classes
	g_list_foreach(classList,processClass,NULL);

	return(result);
}
/* }}} */

/* ruleset functions - static GList *createClassRuleset(struct confClass_t *confClass) {{{ */
// Create a classes ruleset
static GList *createClassRuleset(struct confClass_t *confClass)
{
	GList *addresses = confClass->addresses;
	char *buffer;
	char *tmp;
	GList *result = NULL;
	int include = 1;


	// Function to translate each parameter
	void processParams(gpointer p_key, gpointer p_value, gpointer user_data)
	{
		char *key = (char*) p_key;
		char *value = (char*) p_value;
		int done = 0;
		int this_include = 1;
		
		// Clear temp space
		memset(tmp,'\0',BUFFER_SIZE);
	
		// Processing time...	
		if (!xmlStrcmp(key,"cmd-line") && !done)
		{
			snprintf(tmp,BUFFER_SIZE," %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!xmlStrcmp(key,"dst") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--destination %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!xmlStrcmp(key,"dst-iface") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--out-interface %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!xmlStrcmp(key,"dst-port") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--destination-port %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!xmlStrcmp(key,"name") && !done)  // Ignore this one
		{
			done = 1;
			this_include = 0;
		}
		if (!xmlStrcmp(key,"proto") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--protocol %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!xmlStrcmp(key,"src") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--source %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!xmlStrcmp(key,"src-iface") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--in-interface %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!xmlStrcmp(key,"src-port") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--source-port %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
	
		// Check if we must complain	
		if (done && include)
			include = 1;
		else
			if (done && !this_include)
			{
				fprintf(stderr,"ERROR: Option \"%s\" no recognized\n",key);
			}
	}


	// We must loop with the addresses somewhere	
	void processAddress(gpointer data, gpointer user_data)
	{
		struct confClassAddress_t *address = (struct confClassAddress_t*) data;

		include = 1;
		// Call for each address & process if we must be included
		g_hash_table_foreach(address->params,processParams,NULL);
		if (include)
			result = g_list_append(result,strndup(buffer,BUFFER_SIZE));
		// Blank buffer to avoid overrun fuckup
		memset(buffer,'\0',BUFFER_SIZE);
	}
	

	// Allocate us sum memory
	buffer = (char *) malloc0(BUFFER_SIZE);
	tmp = (char *) malloc(BUFFER_SIZE);

	// And loop with all addresses
	g_list_foreach(addresses,processAddress,NULL);

	// Cleanup...
	free(tmp);
	free(buffer);
	
	// Returning the rule list 
	return result;	
}
/* }}} */

/* ruleset functions - static char *createNATRuleset(GHashTable *properties) {{{ */
// Function which creates our "special" nat extra cmd_line options
static char *createNATRuleset(GHashTable *properties)
{
	char *buffer;
	char *tmp;


	// Function which translates the parameters
	void processParams(gpointer p_key, gpointer p_value, gpointer user_data)
	{
		char *key = (char*) p_key;
		char *value = (char*) p_value;
		int done = 0;
		
		// Clear temp space
		memset(tmp,'\0',BUFFER_SIZE);
	
		// Processing time...	
		if (!xmlStrcmp(key,"to-src") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--to-source %s ",value);
			strncat(buffer,tmp,BUFFER_SIZE);
			done = 1;
		}
		if (!xmlStrcmp(key,"to-dst") && !done)
		{
			snprintf(tmp,BUFFER_SIZE,"--to-destination %s ",value);
			strncat(buffer,tmp,BUFFER_SIZE);
			done = 1;
		}
		if (!xmlStrcmp(key,"name") && !done)  // Ignore this one
			done = 1;
	
		// Check if we must complain	
		if (!done)
		{
			fprintf(stderr,"ERROR: Option \"%s\" no recognized\n",key);
		}
	}

	// Allocate us sum memory
	buffer = (char *) malloc0(BUFFER_SIZE);
	tmp = (char *) malloc(BUFFER_SIZE);

	// And loop with all properties
	g_hash_table_foreach(properties,processParams,NULL);

	// Free sum mem
	free(tmp);
	
	// Return the string FIXME - free mem in parent?
	return buffer;	
}
/* }}} */

/* ruleset functions - static char *createTrafficRuleset(GHashTable *properties) {{{ */
// Function which creates our "special" traffic extra cmd_line options
static char *createTrafficRuleset(GHashTable *properties)
{
	char *buffer;
	char *tmp;


	// Function which translates the parameters
	void processParams(gpointer p_key, gpointer p_value, gpointer user_data)
	{
		char *key = (char*) p_key;
		char *value = (char*) p_value;
		int done = 0;
		
		// Clear temp space
		memset(tmp,'\0',BUFFER_SIZE);
	
		// Processing time...	
		if (!done && !xmlStrcmp(key,"nfmark"))
		{
			snprintf(tmp,BUFFER_SIZE,"MARK --set-mark %s ",value);
			strcat(buffer,tmp);
			done = 1;
		}
		if (!done && !xmlStrcmp(key,"prio"))
			done = 1;
		if (!done && !xmlStrcmp(key,"report-timeout"))
			done = 1;
		if (!done && !xmlStrcmp(key,"burst-rate"))
			done = 1;
		if (!done && !xmlStrcmp(key,"max-rate"))
			done = 1;
		if (!done && !xmlStrcmp(key,"queue-len"))
			done = 1;
		if (!done && !xmlStrcmp(key,"queue-size"))
			done = 1;
		if (!done && !xmlStrcmp(key,"rate"))
			done = 1;
		if (!done && !xmlStrcmp(key,"stats-len"))
			done = 1;
		if (!done && !xmlStrcmp(key,"name"))
			done = 1;
		if (!done && !xmlStrcmp(key,"burst-threshold"))
			done = 1;
	
		// Check if we must complain	
		if (!done)
		{
			fprintf(stderr,"ERROR: Flow option \"%s\" not recognized\n",key);
		}
	}

	// Allocate us sum memory
	buffer = (char *) malloc0(BUFFER_SIZE);
	tmp = (char *) malloc(BUFFER_SIZE);

	// And loop with all properties
	g_hash_table_foreach(properties,processParams,NULL);

	// Free sum mem
	free(tmp);
	
	// Return the string FIXME - free mem in parent?
	return buffer;	
}
/* }}} */

/* parsing functions - static GHashTable *moduleLoadList(char *filename) {{{ */
// Parse the GLOBAL section to get modules we must load
GHashTable *getModuleLoadHash(char *filename)
{
	xmlDocPtr doc;
    	xmlNodePtr cur;
	GHashTable *moduleHash = NULL;

	
	moduleHash = g_hash_table_new_full(g_str_hash,g_str_equal,free,free);

	// COMPAT: Do not genrate nodes for formatting spaces
    	LIBXML_TEST_VERSION
	xmlKeepBlanksDefault(0);
	
	// FIXME - check if file exists
	// Build an XML tree from a the file
	doc = xmlParseFile(filename);
    	if (doc == NULL) 
		return(NULL);

	// Check the document is of the right kind
	cur = xmlDocGetRootElement(doc);
	if (cur == NULL) 
	{
		fprintf(stderr,"ERROR: Empty document\n");
		xmlFreeDoc(doc);
		return(NULL);
	}
	
	// Check if we have the right root element & block
	if (xmlStrcmp(cur->name, (const xmlChar *) "firewall")) 
	{
		fprintf(stderr,"ERROR: Document of the wrong type, root node != firewall");
		xmlFreeDoc(doc);
		return(NULL);
	}


	// Walk the tree.
	cur = cur->xmlChildrenNode;
	while (cur) 
	{
    		xmlNodePtr myNode = cur;
	
		
		// Try find sections
		if (!xmlStrcmp(myNode->name, (const xmlChar *) "global"))
		{
			// We don't care what the top level element name is...
			myNode = myNode->xmlChildrenNode;
			while (myNode)
			{
				int validTag = 0;


				// Check if we found a class
				if (!xmlStrcmp(myNode->name, (const xmlChar *) "class"))
					validTag = 1;

				// Check if we found modules to load
				if (!xmlStrcmp(myNode->name, (const xmlChar *) "modules"))
				{
					xmlNodePtr moduleNode = myNode->xmlChildrenNode;


					while (moduleNode)
					{
						// Check if we are infact doing an address
						if (!xmlStrcmp(moduleNode->name, (const xmlChar *) "load"))
						{
							char *moduleName = xmlGetProp(moduleNode, (const xmlChar *) "name");
							char *params = xmlGetProp(moduleNode, (const xmlChar *) "params");
							int ok = 1;
							

							
							// First check if our name is valid
							if (!moduleName)
							{
								fprintf(stderr,"ERROR: Tag <LOAD /> requires a name\n");
								ok = 0;
							}

							// Then if we a duplicate
							if (ok && g_hash_table_lookup(moduleHash,moduleName))
							{
								fprintf(stderr,"ERROR: Duplicate module %s in MODULES section\n",moduleName);
								ok = 0;
							}
					
							// Finally if it is still ok...	
							if (ok)
							{
								// Check if we can dup params
								if (params)
									params = strdup(params);

								// And insert	
								g_hash_table_insert(moduleHash,moduleName,params);
							}
						}
						else
							fprintf(stderr,"WARNING: Unknown tag %s in MODULES section\n",moduleNode->name);

						moduleNode = moduleNode->next;
					}

					validTag = 1;
				}

				// And a plain text tag...
				if (!xmlStrcmp(myNode->name, (const xmlChar *) "text"))
					validTag = 1;

				// Check if we knew what the tag was
				if (!validTag)
					fprintf(stderr,"WARNING: Unknown tag %s in GLOBAL section\n",myNode->name);
		
				// Advance "class"
				myNode = myNode->next;
			}
		}

				
	   	// Next plz!! 
		cur = cur->next;
	}


	// Clean up everything else before quitting.
	xmlCleanupParser();

	return(moduleHash);
}
/* }}} */

/* parsing functions - static GHashTable *parseGlobal(xmlDocPtr doc, xmlNodePtr cur) {{{ */
// Parse the GLOBAL section
static GHashTable *parseGlobal(xmlDocPtr doc, xmlNodePtr cur)
{
	char *className;
	xmlNodePtr classNode;
	GHashTable *classHash;
	struct confClass_t *tempClass;
	struct confClassAddress_t *tempClassAddress;
	

	classHash = g_hash_table_new(g_str_hash,g_str_equal);

	// We don't care what the top level element name is...
	cur = cur->xmlChildrenNode;
	while (cur)
	{
		int validTag = 0;


		// Check if we found a class
		if (!xmlStrcmp(cur->name, (const xmlChar *) "class"))
		{
			className = xmlGetProp(cur, (const xmlChar *) "name");
			// We have found a class!!
			if (className)
			{
				if (g_hash_table_lookup(classHash,className) == NULL)
				{
					// Allocate our class
					tempClass = (struct confClass_t*) malloc(sizeof(struct confClass_t));
					tempClass->name = className;
					tempClass->addresses = NULL;
				
					// Loop with class contents
					classNode = cur->xmlChildrenNode;
					while (classNode)
					{

						// Check if we are infact doing an address
						if (!xmlStrcmp(classNode->name, (const xmlChar *) "address"))
						{
							// Allocate a new address structure
							tempClassAddress = (struct confClassAddress_t*) malloc(sizeof(struct confClassAddress_t));
							tempClassAddress->params = getProperties(classNode);
							// FIXME - get this things name from the hash above
							tempClassAddress->name = NULL;
							// FIXME - we should parse the attribs here??
						
							tempClass->addresses = g_list_append(tempClass->addresses,tempClassAddress);
						}
						// Advance "address"
						classNode = classNode->next;
					}
				
					// Create our ruleset
					tempClass->ruleset = createClassRuleset(tempClass);
						
					// Add our class to the major list...
					g_hash_table_insert(classHash,className,tempClass);
				}
				else
					fprintf(stderr,"ERROR: Duplicate class found: %s\n",className);
			}
			else
				fprintf(stderr,"ERROR: We found a class without a name\n");
			validTag = 1;
		}

		// Check if we found modules to load
		if (!xmlStrcmp(cur->name, (const xmlChar *) "modules"))
			validTag = 1;

		// And a plain text tag...
		if (!xmlStrcmp(cur->name, (const xmlChar *) "text"))
			validTag = 1;

		// Check if we knew what the tag was
		if (!validTag)
			fprintf(stderr,"WARNING: Unknown tag %s in GLOBAL section\n",cur->name);
		
		// Advance "class"
		cur = cur->next;
	}

	return(classHash);
}
/* }}} */

/* parsing functions - static GHashTable *parseACL(xmlDocPtr doc, xmlNodePtr cur, GHashTable *classHash) {{{ */
// Parse the ACL section
static GHashTable *parseACL(xmlDocPtr doc, xmlNodePtr cur, GHashTable *fwHash, GHashTable *classHash)
{
	char *tableName;
	char *chainName;
	char *ruleName;
	xmlNodePtr tableNode;
	xmlNodePtr chainNode;
	GList *contentList;
	GHashTable *chainHash;
	GList *ruleList;
	struct confACLTable_t *tmpTable; 
	struct confACLChain_t *tmpChain; 
	struct confACLRule_t *tmpRule; 
	GHashTable *tmpProp;


	// Lets start our parsing...
	cur = cur->xmlChildrenNode;
	while (cur)
	{
		// Check if we found a table...
		if (!xmlStrcmp(cur->name, (const xmlChar *) "table"))
		{
			tableName = xmlGetProp(cur, (const xmlChar *) "name");
			
			// Check we have a table name...
			if (tableName)
			{
				// Find table
				tmpTable = lookupTable(fwHash,tableName);
				chainHash = tmpTable->chains;
		
				// Process our table list...
				tableNode = cur->xmlChildrenNode;
				while (tableNode)
				{
				
					// Check if we are in the chain item
					if (!xmlStrcmp(tableNode->name, (const xmlChar *) "chain"))
					{
						tmpProp = getProperties(tableNode);
						chainName = g_hash_table_lookup(tmpProp,"name");
						
						// Check we have a chain name... this we must
						if (chainName)
						{
							// Find chain
							tmpChain = lookupChain(chainHash,chainName);
							tmpChain->defaultTarget = g_hash_table_lookup(tmpProp,"default");
							ruleList = tmpChain->ruleList;
	
							// Process our chain list
							chainNode = tableNode->xmlChildrenNode;
							while (chainNode)
							{
								// Check we are in our rule...
								if (!xmlStrcmp(chainNode->name, (const xmlChar *) "rule"))
								{
									tmpProp = getProperties(chainNode);
									ruleName = g_hash_table_lookup(tmpProp,"name");
									contentList = splitContents(chainNode);

									tmpRule = (struct confACLRule_t*) malloc(sizeof(struct confACLRule_t));
									tmpRule->target = g_hash_table_lookup(tmpProp,"target");
									tmpRule->classList = contentList;
									tmpRule->params = tmpProp;

									// Check if everything is ok
									if (tmpRule->target == NULL)
									{
										fprintf(stderr,"ERROR: All rules MUST have a target!");
										// FIXME - add memory cleanup here
									}
									else
									{
										// Build our ruleset
										tmpChain->ruleset = g_list_concat(tmpChain->ruleset,
													createRuleset(classHash,tmpRule->classList,
													chainName,tmpRule->target,
													g_hash_table_lookup(tmpProp,"cmd-line"),
													NULL)
												);
										
										ruleList = g_list_append(ruleList,tmpRule);
									}
								}
								// Advance chain node
								chainNode = chainNode->next;
							}
						}
					}
					// Advance table	
					tableNode = tableNode->next;
				}
			}
			else
				fprintf(stderr,"ERROR: Table found without a name\n");
		}
		// Advance our acl node
		cur = cur->next;
	}

	return(fwHash);
}
/* }}} */

/* parsing functions - static void parseNAT(xmlDocPtr doc, xmlNodePtr cur, GHashTable *fwHash, GHashTable *classHash) {{{ */
// Parse the NAT section
static void parseNAT(xmlDocPtr doc, xmlNodePtr cur, GHashTable *fwHash, GHashTable *classHash)
{
	char *tableName;
	char *chainName;
	char *ruleName;
	char *target;
	char *cmd_line;
	xmlNodePtr chainNode;
	GList *contentList;
	
	GHashTable *chainHash;
	GList *ruleList;
	struct confACLTable_t *tmpTable; 
	struct confACLChain_t *tmpChain; 
	struct confACLRule_t *tmpRule; 
	GHashTable *tmpProp;


	// Lets start our parsing...
	cur = cur->xmlChildrenNode;
	while (cur)
	{
		tableName = NULL;
		chainName = NULL;
		target = NULL;

		// Check if we found a table...
		if (!xmlStrcmp(cur->name, (const xmlChar *) "snat"))
		{
			tableName = "nat";
			chainName = "POSTROUTING";
			target = "SNAT";
		}	
		if (!xmlStrcmp(cur->name, (const xmlChar *) "dnat"))
		{
			tableName = "nat";
			chainName = "PREROUTING";
			target = "DNAT";
		}	

		// Check we have a table name...
		if (tableName)
		{
			// Find...
			tmpTable = lookupTable(fwHash,tableName);
			chainHash = tmpTable->chains;
		
			// Process our table list...
			chainNode = cur->xmlChildrenNode;
			while (chainNode)
			{
				// Find chain...
				tmpChain = lookupChain(chainHash,chainName);
				ruleList = tmpChain->ruleList;
				tmpChain->defaultTarget = "ACCEPT";
	
				// Check we are in our rule...
				if (!xmlStrcmp(chainNode->name, (const xmlChar *) "rule"))
				{
					tmpProp = getProperties(chainNode);
					ruleName = g_hash_table_lookup(tmpProp,"name");
					contentList = splitContents(chainNode);

					tmpRule = (struct confACLRule_t*) malloc(sizeof(struct confACLRule_t));
					tmpRule->target = target;
					tmpRule->classList = contentList;
					tmpRule->params = tmpProp;

					// Check if everything is ok
					if (tmpRule->target == NULL)
					{
						fprintf(stderr,"ERROR: All rules MUST have a target!");
						// FIXME - add memory cleanup here
					}
					else
					{
						// Get our extra params
						cmd_line = createNATRuleset(tmpProp);	
						// Build our ruleSet
						tmpChain->ruleset = g_list_concat(tmpChain->ruleset,
								createRuleset(classHash,tmpRule->classList,
								chainName,tmpRule->target,NULL,cmd_line));
						// Voila! we got it!
						ruleList = g_list_append(ruleList,tmpRule);
					}
					
					// Advance table	
					chainNode = chainNode->next;
				}
				
			}
		}
		// Advance our NAT node
		cur = cur->next;
	}

}
/* }}} */

/* parsing functions - static void parseTraffic(xmlDocPtr doc, xmlNodePtr cur, GHashTable *fwHash, GHashTable *classHash) {{{ */
// Parse the TRAFFIC section
static void parseTraffic(xmlDocPtr doc, xmlNodePtr cur, GHashTable *fwHash, GHashTable *classHash)
{
	char *tableName = "mangle";
	char *target = "";
	GHashTable *chainHash;
	struct confACLTable_t *tmpTable; 
	struct confACLChain_t *aChain; 


	// Parse a rule... from a node of course
	void parseRule(xmlNodePtr node, char *parent)
	{
		GHashTable *tmpProp;
		char *nfmark;
		struct confACLChain_t *tmpChain; 
		char *chainName;
		char *ruleName;
		GList *contentList;
		struct confACLRule_t *tmpRule; 
		char *cmd_line;
		GList *ruleList;


		// Pull children
		node = node->xmlChildrenNode;
		// Loop
		while (node)
		{
			int validTag = 0;
			

			// Grab our properties
			tmpProp = getProperties(node);
			nfmark = g_hash_table_lookup(tmpProp,"nfmark");
				
			// Check for a flow queue
			if (!xmlStrcmp(node->name, (const xmlChar *) "queue"))
			{
				// Process ourselves
				if (nfmark)
				{
					chainName = "FORWARD";
					
					// Grab chain...
					tmpChain = lookupChain(chainHash,chainName);
					ruleList = tmpChain->ruleList;
					
					ruleName = g_hash_table_lookup(tmpProp,"name");
					contentList = splitContents(node);

					tmpRule = (struct confACLRule_t*) malloc(sizeof(struct confACLRule_t));
					tmpRule->target = target;
					tmpRule->classList = contentList;
					tmpRule->params = tmpProp;

					// Check if everything is ok
					if (tmpRule->target == NULL)
					{
						fprintf(stderr,"ERROR: All rules MUST have a target!");
						// FIXME - add memory cleanup here
					}
					else
					{
						// Build our ruleSet
						cmd_line = createTrafficRuleset(tmpProp);	
						tmpChain->ruleset = g_list_concat(tmpChain->ruleset,
								createRuleset(classHash,tmpRule->classList,
								chainName,tmpRule->target,NULL,cmd_line));
						
						ruleList = g_list_append(ruleList,tmpRule);
					}
				}
				validTag = 1;
			}

			if (!xmlStrcmp(node->name, (const xmlChar *) "flow"))
			{
				// Process ourselves
				if (nfmark)
				{
					chainName = "FORWARD";
					
					// Grab chain...
					tmpChain = lookupChain(chainHash,chainName);
					ruleList = tmpChain->ruleList;
					
					ruleName = g_hash_table_lookup(tmpProp,"name");
					contentList = splitContents(node);

					tmpRule = (struct confACLRule_t*) malloc(sizeof(struct confACLRule_t));
					tmpRule->target = target;
					tmpRule->classList = contentList;
					tmpRule->params = tmpProp;

					// Check if everything is ok
					if (tmpRule->target == NULL)
					{
						fprintf(stderr,"ERROR: All rules MUST have a target!");
						// FIXME - add memory cleanup here
					}
					else
					{
						// Build our ruleSet
						cmd_line = createTrafficRuleset(tmpProp);	
						tmpChain->ruleset = g_list_concat(tmpChain->ruleset,
								createRuleset(classHash,tmpRule->classList,
								chainName,tmpRule->target,NULL,cmd_line));
						
						ruleList = g_list_append(ruleList,tmpRule);
					}
				}
				// Parse child flow... see if its embedded
				parseRule(node,g_hash_table_lookup(tmpProp,"name"));
				validTag = 1;
			}
		
			// Check for the tags we don't really use here...
			if (!xmlStrcmp(node->name, (const xmlChar *) "group"))
				validTag = 1;

			if (!xmlStrcmp(node->name, (const xmlChar *) "text"))
				validTag = 1;

			// Check if the tag we had was valid
			if (!validTag)
				fprintf(stderr,"WARNING: Unknown tag %s\n",node->name);
			
			node = node->next;
		}
	}


	// See if we have a table by this name
	tmpTable = lookupTable(fwHash,"filter");
	// Check if we already did the chain or not
	aChain = lookupChain(tmpTable->chains,"bwmd");
	// Add to our ruleset
	aChain->ruleset = g_list_append(aChain->ruleset,"-A bwmd -m mark ! --mark 0 -j QUEUE");


	tmpTable = lookupTable(fwHash,tableName);
	chainHash = tmpTable->chains;
	
	// Parse our rules...
	parseRule(cur,NULL);
}
/* }}} */

/* firewall functions - GList *createFirewallRules(char *filename) {{{ */
// Function to build a list of rules which iptables-restore takes
GList *createFirewallRules(char *filename) 
{
	xmlDocPtr doc;
    	xmlNodePtr cur;
	GHashTable *classHash = NULL;
	GHashTable *fwHash = NULL;
	GList *result = NULL;
	struct confACLTable_t *tmpTable; 
	struct confACLChain_t *aChain; 
	char *tableList[] = {"filter","mangle",NULL};
	char *tableName;
	int i;


	// Loop with a table	
	void processTable(gpointer p_key, gpointer p_value, gpointer p_user_data)
	{
		char *aRule;
		char *tableName = (char *) p_key;
		struct confACLTable_t *table = (struct confACLTable_t*) p_value;
		
		
		// Create chains...
		void processChains(gpointer p1_key, gpointer p1_value, gpointer p1_user_data)
		{
			struct confACLChain_t *chain = (struct confACLChain_t*) p1_value;

			
			aRule = (char *) malloc0(BUFFER_SIZE);
			// Check if we have a default target for the chain or not		
			if (chain->defaultTarget)	
				snprintf(aRule,BUFFER_SIZE,":%s %s\n",chain->name,chain->defaultTarget);
			else
				snprintf(aRule,BUFFER_SIZE,":%s -\n",chain->name);
			result = g_list_append(result,aRule);		
		}
			
		// Print out rules
		void processRules(gpointer p1_key, gpointer p1_value, gpointer p1_user_data)
		{
			struct confACLChain_t *chain = (struct confACLChain_t*) p1_value;
			
			
			// And each item therein
			void processRuleset(gpointer data, gpointer user_data)
			{
				char *rule = (char *) data;
				
				
				aRule = (char *) malloc0(BUFFER_SIZE);
				snprintf(aRule,BUFFER_SIZE,"%s\n",rule);
				result = g_list_append(result,aRule);
			}
		
			// Loop with all the ruleset items	
			g_list_foreach(chain->ruleset,processRuleset,NULL);
		}

		
		aRule = (char *) malloc0(BUFFER_SIZE);
		snprintf(aRule,BUFFER_SIZE,"*%s\n",tableName);
		result = g_list_append(result,aRule);
		
		// First generate chain names
		g_hash_table_foreach(table->chains,processChains,NULL);
		// Then the rules for the chains
		g_hash_table_foreach(table->chains,processRules,NULL);
		
		aRule = (char *) malloc0(BUFFER_SIZE);
		snprintf(aRule,BUFFER_SIZE,"COMMIT\n");
		result = g_list_append(result,aRule);
	}


	
	// COMPAT: Do not genrate nodes for formatting spaces
    	LIBXML_TEST_VERSION
	xmlKeepBlanksDefault(0);
	
	// FIXME - check if file exists
	// Build an XML tree from a the file
	doc = xmlParseFile(filename);
    	if (doc == NULL) 
		return(NULL);

	// Check the document is of the right kind
	cur = xmlDocGetRootElement(doc);
	if (cur == NULL) 
	{
		fprintf(stderr,"ERROR: Empty document\n");
		xmlFreeDoc(doc);
		return(NULL);
	}
	
	// Check if we have the right root element & block
	if (xmlStrcmp(cur->name, (const xmlChar *) "firewall")) 
	{
		fprintf(stderr,"ERROR: Document of the wrong type, root node != firewall");
		xmlFreeDoc(doc);
		return(NULL);
	}


	// Init everything...	
	fwHash = g_hash_table_new(g_str_hash,g_str_equal);
	i = 0;
	tableName = tableList[i];
	while (tableName)
	{
		// See if we have a table by this name
		tmpTable = lookupTable(fwHash,tableName);
		// Check if we already did the chain or not
		aChain = lookupChain(tmpTable->chains,"INPUT");
		aChain->defaultTarget = "ACCEPT";
		aChain = lookupChain(tmpTable->chains,"FORWARD");
		aChain->defaultTarget = "ACCEPT";
		aChain = lookupChain(tmpTable->chains,"OUTPUT");
		aChain->defaultTarget = "ACCEPT";
		// Advance...
		i++;
		tableName = tableList[i];
	}

	// Walk the tree.
	cur = cur->xmlChildrenNode;
	while (cur) 
	{
		// Try find sections
		if (!xmlStrcmp(cur->name, (const xmlChar *) "global"))
			classHash = parseGlobal(doc,cur);
		if (!xmlStrcmp(cur->name, (const xmlChar *) "acl"))
			parseACL(doc,cur,fwHash,classHash);
		if (!xmlStrcmp(cur->name, (const xmlChar *) "nat"))
			parseNAT(doc,cur,fwHash,classHash);
		if (!xmlStrcmp(cur->name, (const xmlChar *) "traffic"))
			parseTraffic(doc,cur,fwHash,classHash);
				
	   	// Next plz!! 
		cur = cur->next;
	}


	// Clean up everything else before quitting.
	xmlCleanupParser();

	// Build rule list
	g_hash_table_foreach(fwHash,processTable,NULL);	

	return(result);
}
/* }}} */

/* bwmd functions - struct flowData_t createFlowData(char *filename) {{{ */
// Parse the TRAFFIC section and create flows + queues
struct flowData_t createFlowData(char *filename)
{
	xmlDocPtr doc;
    	xmlNodePtr cur;
	GList *pktQueues = NULL;
	GList *flows = NULL, *groups = NULL, *groupItem;
	struct flowData_t ret;


	// Parse a tag... from a node of course
	void parseTag(xmlNodePtr node, struct flow_t *parentFlow)
	{
		struct flow_t *newFlow = NULL;


		// Pull children
		node = node->xmlChildrenNode;
		// Loop
		while (node)
		{
			int validTag = 0;
			GHashTable *tagProperties = getProperties(node);


//			fprintf(stderr,"Tag: %s\n",node->name);

			// We are a flow...
			if (!xmlStrcmp(node->name, (const xmlChar *) "flow"))
			{
				char *flowName = g_hash_table_lookup(tagProperties,"name");
				int ok = 1;


				// Check everything is ok
				if (!flowName)
				{
					fprintf(stderr,"ERROR: Flow missing the \"name\" property!\n");
					ok = 0;
				}

				
				if (ok)
				{
					char *p;
					long int statsLen, maxQueueSize, maxQueueLen, maxRate, burstRate, reportTimeout, nfmark;
					float parent_th;

					
//					fprintf(stderr,"  Flow found: %s\n",flowName);

					// Grab all properties we need
					p = g_hash_table_lookup(tagProperties,"stats-len");
					statsLen = p ? atol(p) : 10;
				
					p = g_hash_table_lookup(tagProperties,"queue-size");
					maxQueueSize = p ? atol(p) : -1;
					
					p = g_hash_table_lookup(tagProperties,"queue-len");
					maxQueueLen = p ? atol(p) : -1;

					p = g_hash_table_lookup(tagProperties,"max-rate");
					maxRate = p ? atol(p) : 0;

					p = g_hash_table_lookup(tagProperties,"burst-rate");
					burstRate = p ? atol(p) : -1;

					p = g_hash_table_lookup(tagProperties,"nfmark");
					nfmark = p ? atol(p) : -1;
					
					p = g_hash_table_lookup(tagProperties,"burst-threshold");
					parent_th = p ? atof(p) : -1;
					
					p = g_hash_table_lookup(tagProperties,"report-timeout");
					reportTimeout = p ? atol(p) : 0;
					if (reportTimeout < 30 && reportTimeout > 0)
					{
						fprintf(stderr,"ERROR: %s - Tag \"report-timeout\" cannot be below 30 (seconds)\n",flowName);
						reportTimeout = 30;
					}

					// Create our flow
					newFlow = createFlow(flowName,parentFlow,statsLen,maxQueueSize,maxQueueLen,maxRate,burstRate,nfmark,
							parent_th,reportTimeout);
					
					flows = g_list_append(flows,newFlow);
				}
				validTag = 1;
			}
	
			// Process if we a group of flows
			if (!xmlStrcmp(node->name, (const xmlChar *) "group"))
			{
				char *groupName = g_hash_table_lookup(tagProperties,"name");
				int ok = 1;


				// Check we have a flow name
				if (!groupName)
				{
					fprintf(stderr,"ERROR: Group missing the \"name\" property!\n");
					ok = 0;
				}
				
				if (ok)
				{
					char *p;
					long int statsLen, reportTimeout;
					GList *flowList;
					struct group_t *newGroup;
					
					
					// Grab all properties we need
					p = g_hash_table_lookup(tagProperties,"stats-len");
					statsLen = p ? atol(p) : 10;
				
					p = g_hash_table_lookup(tagProperties,"report-timeout");
					reportTimeout = p ? atol(p) : 0;
					if (reportTimeout < 30 && reportTimeout > 0)
					{
						fprintf(stderr,"ERROR: Tag \"report-timeout\" cannot be below 30 (seconds)\n");
						reportTimeout = 30;
					}
				
					flowList = splitContents(node);
					// Create our group
					newGroup = createGroup(groupName,statsLen,reportTimeout,flowList);
					groups = g_list_append(groups,newGroup);
				}
				validTag = 1;
			}
			
			// Process if we a queue for a flow	
			if (!xmlStrcmp(node->name, (const xmlChar *) "queue"))
			{
				char *queuePrio = g_hash_table_lookup(tagProperties,"prio");
				char *queueNfmark = g_hash_table_lookup(tagProperties,"nfmark");
				int ok = 1;


				// Check everything is ok
				if (!queuePrio)
				{
					fprintf(stderr,"ERROR: Queue missing the \"prio\" property!\n");
					ok = 0;
				}
				if (!queueNfmark)
				{
					fprintf(stderr,"ERROR: Queue missing the \"nfmark\" property!\n");
					ok = 0;
				}

				
				if (ok)
				{
					long int prio;
					long int nfmark;
					struct pktQueue_t *pktQueue;


					// Grab all properties we need
					prio = atol(queuePrio);
					if (prio >= NUM_PRIO_BANDS || prio < 0)
					{
						fprintf(stderr,"ERROR: Queue priority must be between 0 and 99\n");
						ok = 0;
					}

					nfmark = atol(queueNfmark);
					
		//			fprintf(stderr,"  Flow queue: %li\n", nfmark);

					if (ok)
					{
						// Create our queue
						pktQueue = createPktQueue(prio,nfmark,parentFlow);
						pktQueues = g_list_append(pktQueues,pktQueue);
					}
				}
				validTag = 1;
			}
			
			// Process if we a queue for a flow	
			if (!xmlStrcmp(node->name, (const xmlChar *) "text"))
				validTag = 1;
		
			// Check if we had a valid tag
			if (validTag)	
				// Parse to see if we have children
				parseTag(node,newFlow);
			else
				fprintf(stderr,"WARNING: Unknown tag %s\n",node->name);
			
			// Advance and continue
			node = node->next;
		}
	}


	// Init...
	ret.pktQueues = NULL;
	ret.flows = NULL;
	ret.groups = NULL;
	
	// Compat, set sum stuff
    	LIBXML_TEST_VERSION
	xmlKeepBlanksDefault(0);

	// Build an XML tree from a the file
	doc = xmlParseFile(filename);
    	if (!doc) 
		return(ret);

	// Check the document is of the right kind
	cur = xmlDocGetRootElement(doc);
	if (!cur) 
	{
		fprintf(stderr,"ERROR: Empty document\n");
		xmlFreeDoc(doc);
		return(ret);
	}
	
	// Check if we have the right root element & block
	if (xmlStrcmp(cur->name, (const xmlChar *) "firewall")) 
	{
		fprintf(stderr,"ERROR: Document of the wrong type, root node != firewall");
		xmlFreeDoc(doc);
		return(ret);
	}

	// Walk the tree.
	cur = cur->xmlChildrenNode;
	while (cur) 
	{
		if (!xmlStrcmp(cur->name, (const xmlChar *) "traffic"))
			parseTag(cur,NULL);
	   	// Next plz!! 
		cur = cur->next;
	}

	// Link flow->groups to the actual groups
	groupItem = g_list_first(groups);
	while (groupItem)
	{
		struct group_t *tmpGroup = groupItem->data;
		GList *contentItem = g_list_first(tmpGroup->flowList);

		// Loop with contentList items
		while (contentItem)
		{
			GList *flowItem = g_list_first(flows);
			int found = 0;
			
			// Loop with flow items
			while (flowItem && !found)	
			{
				struct flow_t *tmpFlow = flowItem->data;

				// Compare flow name to the content list inside the group
				if (strcmp((char *) contentItem->data, tmpFlow->flowName) == 0)
				{
					// If it matches, check we arn't listed twice
					if (!g_list_find(tmpFlow->groups,tmpGroup))
					{
						tmpFlow->groups = g_list_append(tmpFlow->groups,tmpGroup);
						found = 1;
					}
					else
						fprintf(stderr,"WARNINGE: Flow %s listed twice in group %s\n",tmpFlow->flowName,tmpGroup->groupName);
				}
				
				flowItem = g_list_next(flowItem);
			}

			if (!found)
				fprintf(stderr,"WARNING: Failed to find flow %s for group %s\n",(char *) contentItem->data,tmpGroup->groupName);
			
			contentItem = g_list_next(contentItem);
		}
		
		groupItem = g_list_next(groupItem);
	}	
	
	// Cleanup stuff
	xmlCleanupParser();

	// Set our return stuff
	ret.pktQueues = pktQueues;
	ret.flows = flows;
	ret.groups = groups;
	
	return(ret);
}
/* }}} */


